#include "RepoEditor.h"
#include "RepoEditorSortFilterProxyModel.h"

RepoEditor::SortFilterProxyModel::SortFilterProxyModel(QObject*  parent)
   : QSortFilterProxyModel(parent)
{
}

bool RepoEditor::SortFilterProxyModel::lessThan(const QModelIndex&  left,
                                                const QModelIndex&  right) const
{
   switch (left.column()) {
   case ID:
   case NAME:
      if (left.model() == NULL)  return false;
      if (right.model() == NULL) return false;
      {
         QVariant  lv = left.model()->data(left,  Qt::DisplayRole);
         QVariant  rv = left.model()->data(right, Qt::DisplayRole);
         
         if (lv.type() != QVariant::String
             || rv.type() != QVariant::String) {
            return QSortFilterProxyModel::lessThan(left, right);
         }

         // The algorithm here basically splits the strings into sequences of
         // non-digits and digits.  It then compares each in turn.  When
         // comparing the non-digit sequences, it compares them lexically.
         // When comparing the digit sequences, it compares them numerically.
         // If we run out of characters on either side, the remaining sequences
         // are treated as "" (non-digits) or 0 (digits).
         // 
         // So,
         //    "foo 1.2.3" < "foo 1.2.4" < "foo 1.2.10" < "foo 1.3" < "fox 0.1"
         //
         // We also see things like:
         //    "foo 1.2.9" < "foo 1.2.n"
         // But this happens because...
         //    "foo " == "foo ", 1 == 1, "." == ".", 2 == 2, "." < "2.n".
         //
         QString  ls = lv.toString();
         QString  rs = rv.toString();

         int  lix = 0;
         int  rix = 0;

         for (;;) {
            // Find digit, then compare non-digit prefix
            {
               int  lterm = lix;
               int  rterm = rix;
               while (lterm < ls.length() && !ls[lterm].isDigit()) lterm++;
               while (rterm < rs.length() && !rs[rterm].isDigit()) rterm++;
               int  cmp = ls.mid(lix, lterm-lix).compare(rs.mid(rix, rterm-rix));
               if (cmp < 0) return true;
               if (cmp > 0) return false;
               lix = lterm;
               rix = rterm;
            }
            if (lix >= ls.length() && rix >= rs.length()) break;

            // Find non-digit, then compare digit prefix
            {
               int  lterm = lix;
               int  rterm = rix;
               while (lterm < ls.length() && ls[lterm].isDigit()) lterm++;
               while (rterm < rs.length() && rs[rterm].isDigit()) rterm++;
               qulonglong  lval = ls.mid(lix, lterm-lix).toULongLong(0, 10);
               qulonglong  rval = rs.mid(rix, rterm-rix).toULongLong(0, 10);
               if (lval < rval) return true;
               if (lval > rval) return false;
               lix = lterm;
               rix = rterm;
            }
            if (lix >= ls.length() && rix >= rs.length()) break;
         }

         // They're equal
         return false;
      }
      break;
   case USE:
   case ENABLED:
   default:
      return QSortFilterProxyModel::lessThan(left, right);
   }
}
